# -*- coding: utf-8 -*-

import cPickle as pickle
import hashlib


class PickleMD5KeyGenerator(object):
    """Cache key generator using Pickle and MD5.

    :param int protocol: Pickle protocol version.
    """

    def __init__(self, protocol=-1):
        self._protocol = protocol

    def __call__(self, args, kwargs):
        pickled_args = pickle.dumps((args, sorted(kwargs.iteritems())),
                               protocol=self._protocol)
        return hashlib.md5(pickled_args).hexdigest()



# -*- coding: utf-8 -*-

from mongo_memoize.key_generator import PickleMD5KeyGenerator

from nose.tools import *


def test_pickle_md5_key_generator():
    ins = PickleMD5KeyGenerator()
    key1 = ins((1, 2), dict(kwarg1=1))
    ok_(isinstance(key1, basestring))

    key2 = ins((1,), dict(kwarg1=1))
    ok_(key1 != key2)

    key3 = ins((1, 2), dict(kwarg1=1, kwarg2=2))
    ok_(key1 != key3)



# -*- coding: utf-8 -*-

from mongo_memoize.serializer import NoopSerializer, PickleSerializer

from nose.tools import *


def test_noop_serializer():
    ins = NoopSerializer()
    obj = dict(a=1, b=2)

    serialized = ins.serialize(obj)
    eq_(obj, serialized)

    deserialized = ins.deserialize(serialized)
    eq_(obj, deserialized)


def test_pickle_serializer():
    ins = PickleSerializer()
    obj = dict(a=1, b=2)

    serialized = ins.serialize(obj)
    ok_(isinstance(serialized, basestring))

    deserialized = ins.deserialize(serialized)
    eq_(obj, deserialized)



# -*- coding: utf-8 -*-

import pymongo
from functools import wraps

from key_generator import PickleMD5KeyGenerator
from serializer import PickleSerializer


def memoize(
    db_name='mongo_memoize', host='localhost', port=27017, collection_name=None,
    prefix='memoize_', capped=False, capped_size=100000000, capped_max=None,
    connection_options={}, key_generator=None, serializer=None
):
    """A decorator that caches results of the function in MongoDB.

    Usage:

        >>> from mongo_memoize import memoize
        >>> @memoize()
        ... def some_function():
        ...     pass
        ...

    :param str db_name: MongoDB database name.
    :param str host: MongoDB host name.
    :param int port: MongoDB port.
    :param str collection_name: MongoDB collection name. If not specified, the
        collection name is generated automatically using the prefix, the module
        name, and the function name.
    :param str prefix: Prefix of the MongoDB collection name. This argument is
        only valid when the collection_name argument is not specified.
    :param bool capped: Whether to use the capped collection.
    :param int capped_size: The maximum size of the capped collection in bytes.
    :param int capped_max: The maximum number of items in the capped collection.
    :param dict connection_options: Additional parameters for establishing
        MongoDB connection.
    :param key_generator: Key generator instance.
        :class:`PickleMD5KeyGenerator <mongo_memoize.PickleMD5KeyGenerator>` is used by default.
    :param serializer: Serializer instance.
        :class:`PickleSerializer <mongo_memoize.PickleSerializer>` is used by default.
    """
    if not serializer:
        serializer  = PickleSerializer()

    if not key_generator:
        key_generator = PickleMD5KeyGenerator()

    def decorator(func):
        db_conn = pymongo.Connection(host, port, *connection_options)[db_name]

        if collection_name:
            col_name = collection_name
        else:
            col_name = '%s%s.%s' % (prefix, func.__module__, func.__name__)

        if capped:
            if col_name not in db_conn.collection_names():
                assert capped_size > 0, 'The size of the capped collection is required.'

                capped_args = {}
                capped_args['size'] = capped_size
                if capped_max:
                    capped_args['max'] = capped_max

                db_conn.create_collection(col_name, capped=True, **capped_args)

        cache_col = db_conn[col_name]
        cache_col.ensure_index('key', unique=True)

        @wraps(func)
        def wrapped_func(*args, **kwargs):
            cache_key = key_generator(args, kwargs)

            cached_obj = cache_col.find_one(dict(key=cache_key))
            if cached_obj:
                return serializer.deserialize(cached_obj['result'])

            ret = func(*args, **kwargs)
            cache_col.update(
                {'key': cache_key},
                {
                    '$set': {
                        'result': serializer.serialize(ret),
                    }
                },
                upsert=True
            )

            return ret

        return wrapped_func

    return decorator



# -*- coding: utf-8 -*-

import base64
import cPickle as pickle


class NoopSerializer(object):
    """Serializer that does nothing.

    .. note::
        It is required that the result of the function can be stored
        directly in MongoDB when you use this serializer.
    """

    def serialize(self, obj):
        return obj

    def deserialize(self, serialized):
        return serialized


class PickleSerializer(object):
    """Serializer using Pickle.

    :param int protocol: Pickle protocol version.
    """
    def __init__(self, protocol=-1):
        self._protocol = protocol

    def serialize(self, obj):
        return base64.b64encode(pickle.dumps(obj, protocol=self._protocol))

    def deserialize(self, serialized):
        return pickle.loads(base64.b64decode(serialized))



# -*- coding: utf-8 -*-

import pymongo
import time
import uuid
from collections import defaultdict

from mongo_memoize.decorator import memoize

from nose.tools import *

call_count = defaultdict(int)
db_name = uuid.uuid1().hex
db_conn = pymongo.Connection()


def tearDown():
    db_conn.drop_database(db_name)


@memoize(db_name=db_name)
def memoized_func():
    call_count['memoized_func'] += 1
    return True


@memoize(db_name=db_name, collection_name='capped_col', capped=True,
         capped_size=1000, capped_max=10)
def memoized_func_capped():
    call_count['memoized_func_capped'] += 1
    return True


def test_memoize():
    ok_(memoized_func())  # should be called
    time.sleep(0.1)
    ok_(memoized_func())
    eq_(1, call_count['memoized_func'])


def test_memoize_capped():
    ok_(memoized_func_capped())  # should be called
    time.sleep(0.1)
    ok_(memoized_func_capped())
    eq_(1, call_count['memoized_func_capped'])

    col_options = db_conn[db_name]['capped_col'].options()
    eq_(True, col_options['capped'])
    eq_(1000.0, col_options['size'])
    eq_(10, col_options['max'])



# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

setup(
    name='mongo-memoize',
    version='0.0.3',
    description='A Python decorator library for caching function results in MongoDB',
    long_description=open('README.rst').read(),
    author='Ikuya Yamada',
    author_email='ikuya@ikuya.net',
    url='http://github.com/ikuyamada/mongo-memoize/',
    packages=find_packages(),
    license=open('LICENSE').read(),
    include_package_data=True,
    keywords=[],
    classifiers=(
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
    ),
    install_requires=[
        'pymongo',
    ],
    tests_require=[
        'nose',
        'mock',
    ],
    test_suite = 'nose.collector'
)
